import unittest
import recurly
import socket
from recurly import Resource
from recurly.resource import Empty
from .mock_resources import MyResource, MySubResource
from .mock_client import MockClient
import unittest.mock as mock
from unittest.mock import Mock, MagicMock


def delete_resource_client():
    conn = MagicMock()
    conn.request = MagicMock(return_value=None)
    response = MagicMock()
    # empty response
    response.status = 204
    response.read.return_value = bytes("", "UTF-8")
    conn.getresponse = MagicMock(return_value=response)
    return mock.patch("http.client.HTTPSConnection", return_value=conn)


def update_resource_client(success):
    conn = MagicMock()
    conn.request = MagicMock(return_value=None)
    response = MagicMock()
    if success:
        response.status = 201
        response.read.return_value = bytes(
            """
            {
                "object": "my_resource", "my_int": 123
            }
            """,
            "UTF-8",
        )
    else:
        response.status = 422
        response.read.return_value = bytes(
            """
            {
                "error": {
                    "type": "validation",
                    "message": "Year must be greater than 2000, Number is required",
                    "params":[{"param":"year","message":"must be greater than 2000"},{"param":"number","message":"is required"}]
                }
            }
            """,
            "UTF-8",
        )

    conn.getresponse = MagicMock(return_value=response)

    return mock.patch("http.client.HTTPSConnection", return_value=conn)


def get_socket_error_client():
    conn = MagicMock()
    conn.request = MagicMock(return_value=None)
    conn.getresponse = MagicMock(side_effect=socket.error("socket failure"))
    return mock.patch("http.client.HTTPSConnection", return_value=conn)


def get_resource_client(success):
    conn = MagicMock()
    conn.request = MagicMock(return_value=None)
    response = MagicMock()
    if success:
        response.status = 200
        response.read.return_value = bytes(
            """
            {
                "object": "my_resource", "my_int": 123
            }
            """,
            "UTF-8",
        )
    else:
        response.status = 404
        response.read.return_value = bytes(
            """
            {
                "error": {
                   "type": "not_found",
                   "message": "Couldn't find Resource with id = 123",
                   "params": [{"param": "resource_id"}]
               }
            }
            """,
            "UTF-8",
        )

    conn.getresponse = MagicMock(return_value=response)

    return mock.patch("http.client.HTTPSConnection", return_value=conn)


expected_headers = {
    "Authorization": "Basic YXBpa2V5Og==",
    "Accept": "application/vnd.recurly.v2018-08-09",
    "Content-Type": "application/json",
}


class TestBaseClient(unittest.TestCase):
    def test_api_version(self):
        client = MockClient("apikey")
        self.assertEqual(client.api_version(), "v2018-08-09")

    def test_successful_GET_200(self):
        with get_resource_client(True) as conn:
            client = MockClient("apikey")
            resource = client.get_resource("123", {"q": 123})
            # conn.request.assert_called_with(
            #     "GET", "/resources/123?q=123", None, headers=expected_headers
            # )
            self.assertEqual(type(resource), MyResource)

    def test_failure_GET_404(self):
        with get_resource_client(False) as conn:
            client = MockClient("apikey")
            with self.assertRaises(recurly.errors.NotFoundError) as e:
                resource = client.get_resource("123")

            # conn.request.assert_called_with(
            #     "GET", "/resources/123?q=123", None, headers=expected_headers
            # )
            err = e.exception.error
            self.assertEqual(err.type, "not_found")

    def test_successful_PUT_201(self):
        with update_resource_client(True) as conn:
            client = MockClient("apikey")
            resource = client.update_resource("123", {"my_int": 123})
            # conn.request.assert_called_with(
            #     "GET",
            #     "/resources/123?q=123",
            #     """{"my_int", 123}""",
            #     headers=expected_headers,
            # )
            self.assertEqual(type(resource), MyResource)
            self.assertEqual(resource.my_int, 123)

    def test_failure_PUT_422(self):
        with update_resource_client(False) as conn:
            client = MockClient("apikey")
            with self.assertRaises(recurly.errors.ValidationError) as e:
                resource = client.update_resource("123", {"my_int": 123})

            # conn.request.assert_called_with(
            #     "GET",
            #     "/resources/123?q=123",
            #     """{"my_int", 123}""",
            #     headers=expected_headers,
            # )
            err = e.exception.error
            self.assertEqual(err.type, "validation")

    def test_DELETE_204(self):
        with delete_resource_client() as conn:
            client = MockClient("apikey")
            resource = client.delete_resource("123")
            # conn.request.assert_called_with(
            #     "DELETE",
            #     "/resources/123",
            #     None,
            #     headers=expected_headers,
            # )
            self.assertIsInstance(resource, Empty)

    def test_failure_socket_error(self):
        with get_socket_error_client() as conn:
            client = MockClient("apikey")
            with self.assertRaises(recurly.NetworkError) as e:
                resource = client.update_resource("123", {"my_int": 123})



class ItemIterator:
    def __init__(self, client, next_url, params):
        self.__page_iterator = PageIterator(client, next_url, params)
        self.__index = 0
        self.__data = None

    def __iter__(self):
        return self

    def __next__(self):
        # If we are in a page, get the next item
        if self.__data and self.__index < len(self.__data) - 1:
            self.__index += 1
        # else we want to fetch a new page and point to the first item
        else:
            self.__data = next(self.__page_iterator)
            self.__index = 0

        if len(self.__data) > 0:
            return self.__data[self.__index]
        else:
            raise StopIteration


class PageIterator:
    def __init__(self, client, next_url, params):
        self.__client = client
        self.__next_url = next_url
        self.__params = params
        self.__has_more = True

    def __iter__(self):
        return self

    def __next__(self):
        if self.__has_more:
            page = self.__client._make_request(
                "GET", self.__next_url, None, self.__params
            )

            # We don't need the params anymore as they are
            # automatically set in __next_url
            if self.__params:
                self.__params = None

            self.__next_url = page.next
            self.__has_more = page.has_more
            return page.data
        else:
            raise StopIteration


class Pager:
    def __init__(self, client, path, params):
        self.__client = client
        self.__path = path
        self.__params = self.__map_array_params(params)

    def pages(self):
        """An iterator that enumerates each page of results."""
        return PageIterator(self.__client, self.__path, self.__params)

    def items(self):
        """An iterator that enumerates each item on the server and paginates
        under the hood.
        """
        return ItemIterator(self.__client, self.__path, self.__params)

    def first(self):
        """Performs a request with the pager `limit` set to 1 and only returns the
        first result in the response.
        """
        params = self.__params.copy()
        params.update({"limit": 1})
        items = ItemIterator(self.__client, self.__path, params)
        try:
            return next(items)
        except StopIteration:
            return None

    def count(self):
        """Makes a HEAD request to the API to determine how many total records exist.
        """
        resource = self.__client._make_request("HEAD", self.__path, None, self.__params)
        return resource.get_response().total_records

    def __map_array_params(self, params):
        """Converts array parameters to CSV strings to maintain consistency with
        how the server expects the request to be formatted while providing the
        developer with an array type to maintain developer happiness!
        """
        return {k: ",".join(v) if isinstance(v, list) else v for k, v in params.items()}



name = "recurly"

import os
from os.path import dirname, basename, isfile
import glob
import ssl
import sys

__version__ = "3.5.1"
__python_version__ = ".".join(map(str, sys.version_info[:3]))

USER_AGENT = "Recurly/{}; python {}; {}".format(
    __version__, __python_version__, ssl.OPENSSL_VERSION
)

# Running in strict mode will throw exceptions
# when API responses don't line up with the client's expectations.
# The client's default behavior is to try to recover from these
# errors. This is used to catch bugs in testing.
# You do not want to enable this for production.
STRICT_MODE = os.getenv("RECURLY_STRICT_MODE", "FALSE").upper() == "TRUE"


class RecurlyError(Exception):
    pass


class ApiError(RecurlyError):
    def __init__(self, message, error):
        super(ApiError, self).__init__(message)
        self.error = error


class NetworkError(RecurlyError):
    pass


from . import errors
from .client import Client
from .resource import Resource
from .response import Response
from .request import Request



#
# This file is automatically created by Recurly's OpenAPI generation process
# and thus any edits you make by hand will be lost. If you wish to make a
# change to this file, please create a Github issue explaining the changes you
# need and we will usher them to the appropriate places.
import recurly


class BadRequestError(recurly.ApiError):
    pass


class InternalServerError(recurly.ApiError):
    pass


class ImmutableSubscriptionError(recurly.ApiError):
    pass


class InvalidApiKeyError(recurly.ApiError):
    pass


class InvalidApiVersionError(recurly.ApiError):
    pass


class InvalidContentTypeError(recurly.ApiError):
    pass


class InvalidPermissionsError(recurly.ApiError):
    pass


class InvalidTokenError(recurly.ApiError):
    pass


class NotFoundError(recurly.ApiError):
    pass


class SimultaneousRequestError(recurly.ApiError):
    pass


class TransactionError(recurly.ApiError):
    pass


class UnauthorizedError(recurly.ApiError):
    pass


class UnavailableInApiVersionError(recurly.ApiError):
    pass


class UnknownApiVersionError(recurly.ApiError):
    pass


class ValidationError(recurly.ApiError):
    pass


class MissingFeatureError(recurly.ApiError):
    pass


class RateLimitedError(recurly.ApiError):
    pass
